#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* include this first, before NO_IMPORT_PYGOBJECT is defined */
#include <Python.h>
#include <structmember.h>

#include <pygobject.h>


#include <Python.h>
#include <structmember.h>

#if (PY_VERSION_HEX < 0x02050000)
typedef int Py_ssize_t;
typedef inquiry lenfunc;
typedef intargfunc ssizeargfunc;
typedef intintargfunc ssizessizeargfunc;
typedef intobjargproc ssizeobjargproc;
typedef intintobjargproc ssizessizeobjargproc;
typedef getreadbufferproc readbufferproc;
typedef getwritebufferproc writebufferproc;
typedef getsegcountproc segcountproc;
#endif

#if PY_VERSION_HEX < 0x02030000
static PyObject *
PyObject_SelfIter(PyObject *obj)
{
   Py_INCREF(obj);
   return obj;
}
#endif

void hst_register_classes (PyObject *d);
void hst_add_constants(PyObject *module, const gchar *strip_prefix);
void gtkhst_register_class(PyObject *dict,
			   const char *type_name,
			   PyTypeObject *type,
			   PyObject *bases);


typedef double BufferItem;
typedef struct  {
      PyObject_HEAD
      int    own;
      int    nitems;
      size_t len;
      BufferItem  *item;    // memory buffer
} gtkhst_buffer;



#if PY_VERSION_HEX < 0x2030000
static PyObject *
PyObject_SelfIter(PyObject *obj)
{
   Py_INCREF(obj);
   return obj;
}
#endif


static PyObject *
gtkhst_buffer_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
   gtkhst_buffer *self;
   self = (gtkhst_buffer *)type->tp_alloc(type,0);
   if (!self)
      return (PyObject *)0;

   
   if (self!=NULL) {
      self->own  = 0;
      self->nitems=0;
      self->len   =0;
      self->item  =0;
   }
   return (PyObject *)self;
}

static  int gtkhst_buffer_init(gtkhst_buffer *self, PyObject *args, PyObject *kwds) {
   PyObject *in=NULL;
   
   // Check that we do not get any keyword argument
   if (kwds != NULL) {
      int i = PyObject_Length(kwds);
      if (i < 0)
	 return -1;
      else if (i > 0) {
	  PyErr_SetString(PyExc_ValueError,
			  "gtkhst.Buffer: gtkhst.Buffer constructor takes "
			  "no keyword arguments");
	  return -1;
      }
   }

   // Get the Object 
   if (!PyArg_ParseTuple(args,"O:gtkhst.Buffer.__init__",&in))
      return -1;

   if (in==NULL) {
       PyErr_SetString(PyExc_ValueError,
		      "gtkhst.Buffer is getting a NULL object to clone"
	 );
      return -1;
   }

   // Get the size and the memory ptr of the buffer
   Py_ssize_t sz;
   char *ptr=0;
   if ( PyObject_AsWriteBuffer(in,(void **)&ptr,&sz) ) {
       PyErr_SetString(PyExc_ValueError, "gtkhst.Buffer: import object is not a buffer");
	 return -1;
   }
   else
       self->own = 0;

   if (!ptr) {
       PyErr_SetString(PyExc_ValueError,"gtkhst.Buffer: NULL pointer");
       return -1;
   }

   if ( sz%sizeof(BufferItem) ) {
       PyErr_SetString(PyExc_ValueError,"gtkhst.Buffer: buffer not aligned");
       return -1;
   }
   
   self->len=sz;
   self->item=(BufferItem *)ptr;
   self->nitems = sz/sizeof(unsigned long);
   return 0;
}



static void
gtkhst_buffer_dealloc(gtkhst_buffer* self)
{
   if (self->own && self->item) 
      PyMem_Free(self->item);
   self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
gtkhst_buffer_tostring(gtkhst_buffer *self, PyObject *unused)
{
    return PyString_FromStringAndSize((char *)self->item,self->len);
}



static PyObject *gtkhst_buffer_get_item(gtkhst_buffer *self, int i) {
   PyObject *out=NULL;
   if ( i>=0 && i<self->nitems) {
       out = PyFloat_FromDouble( self->item[i] );
   }
   if ( out==NULL ) {
       PyErr_SetString(PyExc_IndexError,
		       "gtkhst.Buffer __getitem__ index out of range"
	   );
   }
   return out;
}


static int gtkhst_buffer_length(gtkhst_buffer *self) {
   return self->nitems;
}

static PyObject *gtkhst_buffer_concat(gtkhst_buffer *self, PyObject *o) {
   Py_INCREF(self);
   return (PyObject *)self;
}

static int gtkhst_buffer_ass_item(gtkhst_buffer *self, int i, PyObject *v) {
   if ( i<0 || i>=self->nitems) {
      PyErr_SetString(PyExc_IndexError,
		      "gtkhst.Buffer assignment index out of range");
      return -1;
   }
   if (v==NULL) {
      PyErr_SetString(PyExc_ValueError,
		      "gtkhst.Buffer assignment gets a NULL object");
      return -1;
   }
   if (!PyFloat_Check(v)) {
       PyErr_SetString(PyExc_TypeError,
		       "gtkhst.Buffer: tried to assign a non integer");
       return -1;
   }
   self->item[i] = PyFloat_AsDouble(v);
   return 0;
}

static int gtkhst_buffer_ass_slice(gtkhst_buffer *self, int ilow, int ihigh, PyObject *v) {
    PyErr_SetString(PyExc_SystemError,"gtkhst.Buffer: not implemented");
    return -1;
}

static PySequenceMethods gtkhst_buffer_as_sequence = {
   (lenfunc)gtkhst_buffer_length,           /*sq_length*/
   (binaryfunc)gtkhst_buffer_concat,        /*sq_concat*/
   (ssizeargfunc)0,		           /*sq_repeat*/
   (ssizeargfunc)gtkhst_buffer_get_item,	   /*sq_item*/
   (ssizessizeargfunc)gtkhst_buffer_concat,	   /*sq_slice*/
   (ssizeobjargproc)gtkhst_buffer_ass_item,   /*sq_ass_item*/
   (ssizessizeobjargproc)gtkhst_buffer_ass_slice,/*sq_ass_slice*/
   (objobjproc)0,		           /*sq_contains*/
   (binaryfunc)0,	                   /*sq_inplace_concat*/
   (ssizeargfunc)0	                   /*sq_inplace_repeat*/
};

static int
gtkhst_buffer_getreadbuf(gtkhst_buffer *self, int index, const void **ptr)
{
   if ( index != 0 ) {
      PyErr_SetString(PyExc_SystemError,"Accessing non-existent array segment");
      return -1;
   }
   *ptr = (void *)self->item;
   return self->len;
}

static int
gtkhst_buffer_getwritebuf(gtkhst_buffer *self, int index, const void **ptr) {
   if ( index != 0 ) {
      PyErr_SetString(PyExc_SystemError,
		      "Accessing non-existent array segment");
      return -1;
   }
   *ptr = (void *)self->item;
   return self->len;
}

static int
gtkhst_buffer_getsegcount(gtkhst_buffer *self, int *lenp) {
   if ( lenp )
      *lenp = self->len;
   return 1;
}


static PyBufferProcs gtkhst_buffer_as_buffer = {
	(readbufferproc)gtkhst_buffer_getreadbuf,
	(writebufferproc)gtkhst_buffer_getwritebuf,
	(segcountproc)gtkhst_buffer_getsegcount,
};



static PyObject *gtkhst_buffer_get_nitems(gtkhst_buffer *self, void *closure) {
   return PyInt_FromLong((long)self->nitems);
}

static PyObject *gtkhst_buffer_get_size(gtkhst_buffer *self, void *closure) {
   return PyInt_FromLong((long)self->len);
}

static PyGetSetDef gtkhst_buffer_getsets [] = {
   {"nitems",(getter)gtkhst_buffer_get_nitems,NULL,"the number of items in the array"},
   {"size",(getter)gtkhst_buffer_get_size,NULL,"the size, in bytes, of the buffer"},
   {NULL}
};

static PyObject *
gtkhst_buffer_info(gtkhst_buffer *self, PyObject *unused)
{
   PyObject* retval = NULL;
   retval = PyTuple_New(2);
   if (!retval)
      return NULL;

   PyTuple_SET_ITEM(retval, 0, PyLong_FromVoidPtr(self->item));
   PyTuple_SET_ITEM(retval, 1, PyInt_FromLong((long)(self->nitems)));
   
   return retval;
}

static PyObject *gtkhst_buffer_tolist(gtkhst_buffer* self, PyObject *unused) {
   PyObject *list = PyList_New(self->nitems);
   int i;

   if (list==NULL)
      return NULL;

   for (i=0;i<self->nitems;i++) {
      PyObject *v = gtkhst_buffer_get_item(self, i);
      if (v==NULL) {
	 Py_DECREF(list);
	 return NULL;
      }
      PyList_SetItem(list,i,v);
   }
   return list;
}


static PyObject *
gtkhst_buffer_repr(gtkhst_buffer *a) {
   PyObject *s, *t, *v = NULL;
   int len;

   len = a->nitems;
   if (len == 0) {
      return PyString_FromString("gtkhst.Buffer");
   }
		
   v = gtkhst_buffer_tolist(a, NULL);

   t = PyObject_Repr(v);
   Py_XDECREF(v);
   
   s = PyString_FromString("gtkhst.Buffer (");
   PyString_ConcatAndDel(&s, t);
   PyString_ConcatAndDel(&s, PyString_FromString(")"));
   return s;
}


PyMethodDef gtkhst_buffer_methods[] = {
   {"buffer_info", (PyCFunction)gtkhst_buffer_info, METH_NOARGS,"."},
   {"tolist",	(PyCFunction)gtkhst_buffer_tolist, METH_NOARGS,"."},
   {NULL,		NULL}		/* sentinel */
};
static PyObject *gtkhst_buffer_iter(gtkhst_buffer *self);


static PyTypeObject DBuffertype = {
	PyObject_HEAD_INIT(NULL)
	0,
	"gtkhst.Buffer",
	sizeof(gtkhst_buffer),
	0,
	(destructor)gtkhst_buffer_dealloc,		/* tp_dealloc */
	0,					/* tp_print */
	0,					/* tp_getattr */
	0,					/* tp_setattr */
	0,					/* tp_compare */
	(reprfunc)gtkhst_buffer_repr,			/* tp_repr */
	0,					/* tp_as _number*/
	&gtkhst_buffer_as_sequence,			/* tp_as _sequence*/
	0,			/* tp_as _mapping*/
	0, 					/* tp_hash */
	0,					/* tp_call */
	0,					/* tp_str */
	PyObject_GenericGetAttr,		/* tp_getattro */
	0,					/* tp_setattro */
	&gtkhst_buffer_as_buffer,			/* tp_as_buffer*/
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,  /* tp_flags */
	"Memory array wrapper",				/* tp_doc */
 	0,					/* tp_traverse */
	0,					/* tp_clear */
	0,			/* tp_richcompare */
	0,	                            /* tp_weaklistoffset */
	(getiterfunc)gtkhst_buffer_iter,		/* tp_iter */
	0,					/* tp_iternext */
	gtkhst_buffer_methods,				/* tp_methods */
	0,					/* tp_members */
	gtkhst_buffer_getsets,				/* tp_getset */
	0,					/* tp_base */
	0,					/* tp_dict */
	0,					/* tp_descr_get */
	0,					/* tp_descr_set */
	0,					/* tp_dictoffset */
	(initproc)gtkhst_buffer_init,                      /* tp_init */
	0,			/* tp_alloc */
	gtkhst_buffer_new,				/* tp_new */
};

/*-----------------------------------------------------------
 * The Iterator
 *-----------------------------------------------------------*/

typedef struct {
    PyObject_HEAD
    long			index;
    gtkhst_buffer		*buffer;
    PyObject		* (*getitem)(gtkhst_buffer *, int);
} gtkhst_buffer_iter_object;

static PyObject *
gtkhst_buffer_iter_next(gtkhst_buffer_iter_object *it)
{
//	assert(PyArrayIter_Check(it));
	if (it->index < it->buffer->nitems)
		return (*it->getitem)(it->buffer, it->index++);
	return NULL;
}

static void
gtkhst_buffer_iter_dealloc(gtkhst_buffer_iter_object *it)
{
	PyObject_GC_UnTrack(it);
	Py_XDECREF(it->buffer);
	PyObject_GC_Del(it);
}

static int
gtkhst_buffer_iter_traverse(gtkhst_buffer_iter_object *it, visitproc visit, void *arg)
{
	if (it->buffer != NULL)
		return visit((PyObject *)(it->buffer), arg);
	return 0;
}

static PyTypeObject PyGtkhstBufferIter_Type = {
   PyObject_HEAD_INIT(NULL)
   0,                                      /* ob_size */
   "arrayiterator",                        /* tp_name */
   sizeof(gtkhst_buffer_iter_object),      /* tp_basicsize */
   0,                                      /* tp_itemsize */
   /* methods */
   (destructor)gtkhst_buffer_iter_dealloc,		/* tp_dealloc */
   0,                                      /* tp_print */
   0,                                      /* tp_getattr */
   0,                                      /* tp_setattr */
   0,                                      /* tp_compare */
   0,                                      /* tp_repr */
   0,                                      /* tp_as_number */
   0,                                      /* tp_as_sequence */
   0,                                      /* tp_as_mapping */
   0,                                      /* tp_hash */
   0,                                      /* tp_call */
   0,                                      /* tp_str */
   PyObject_GenericGetAttr,                /* tp_getattro */
   0,                                      /* tp_setattro */
   0,                                      /* tp_as_buffer */
   Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,/* tp_flags */
   0,                                      /* tp_doc */
   (traverseproc)gtkhst_buffer_iter_traverse,	/* tp_traverse */
   0,					/* tp_clear */
   0,                                      /* tp_richcompare */
   0,                                      /* tp_weaklistoffset */
   PyObject_SelfIter,			/* tp_iter */
   (iternextfunc)gtkhst_buffer_iter_next,		/* tp_iternext */
   0,					/* tp_methods */
};

static PyObject *gtkhst_buffer_iter(gtkhst_buffer *self) {
   gtkhst_buffer_iter_object *it;
   it = PyObject_GC_New(gtkhst_buffer_iter_object, &PyGtkhstBufferIter_Type);
   if (it == NULL)
      return NULL;

   Py_INCREF(self);
   it->buffer  = self;
   it->index   = 0;
   it->getitem = gtkhst_buffer_get_item;
   PyObject_GC_Track(it);
   return (PyObject *)it;
}

PyObject *PyGtkhst_Buffer_New(int sz, char *ptr) {
   int own=0;
   if (ptr==0 && sz>0 ) {
       own=1;
       ptr = (char *)PyMem_Malloc(sz);
       if (ptr==NULL) {
	   return PyErr_NoMemory();
       }
   }
   gtkhst_buffer *self = PyObject_New(gtkhst_buffer,&DBuffertype);
   self->own = own;
   self->len = sz;
   self->item=(BufferItem *)ptr;
   self->nitems = sz/sizeof(BufferItem);
   return (PyObject *)self;
}


extern PyMethodDef hst_functions[];


DL_EXPORT(void)
inithst(void)
{
    PyObject *m, *d;
	
    init_pygobject ();

    m = Py_InitModule ("hst", hst_functions);
    d = PyModule_GetDict (m);
	
    hst_register_classes (d);
    hst_add_constants(m,"HST_");
    gtkhst_register_class(d,"Buffer",
			  &DBuffertype,
			  0);

    if (PyErr_Occurred ()) {
        Py_FatalError ("can't initialise module hst");
    }
    
}

void gtkhst_register_class(PyObject *dict,
			const char *type_name,
			PyTypeObject *type,
			PyObject *bases)
{
   const char *s,*class_name;

   class_name=type->tp_name;
   s = strrchr(class_name,'.');
   if (s!=NULL)
      class_name = s+1;

   type->ob_type = &PyType_Type;
   if (bases) {
      type->tp_bases = bases;
      type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
   }
   if (PyType_Ready(type) < 0) {
       printf( "couldn't make the type %s ready\n" ,type->tp_name);
      return;
   }
   
   
   PyDict_SetItemString(dict, (char *)class_name, (PyObject *)type);
}
